==> hgraph-0.2.41-py3-none-any.whl: hgraph/nodes/_tsl_operators.py <==
from typing import Type

from hgraph import compute_node, TSL, TIME_SERIES_TYPE, SIZE, SCALAR, TS, graph, AUTO_RESOLVE, NUMBER, REF, TSD, \
    union, TSS, KEYABLE_SCALAR, TSS_OUT, PythonSetDelta, add_, sub_, mul_, div_, floordiv_, mod_, pow_, lshift_, \
    rshift_, bit_and, bit_or, bit_xor, eq_, ne_, not_, neg_, pos_, invert_, abs_, min_, max_, reduce, zero, \
    str_, PythonTimeSeriesReference, len_, sum_, getitem_
from hgraph._operators._control import merge, all_
from hgraph.adaptors.data_frame._data_source_generators import SIZE_1
from hgraph.nodes import const

__all__ = ("flatten_tsl_values", "tsl_to_tsd", "index_of")


@compute_node
def flatten_tsl_values(tsl: TSL[TIME_SERIES_TYPE, SIZE], all_valid: bool = False) -> TS[tuple[SCALAR, ...]]:
    """
    This will convert the TSL into a time-series of tuples. The value will be the value type of the time-series
    provided to the TSL. If the value has not ticked yet, the value will be None.
    The output type must be defined by the user.

    Usage:
    ```Python
        tsl: TSL[TS[float], Size[3]] = ...
        out = flatten_tsl[SCALAR: float](tsl)
    ```
    """
    return tsl.value if not all_valid or tsl.all_valid else None


@compute_node(overloads=merge)
def merge_default(*tsl: TSL[TIME_SERIES_TYPE, SIZE]) -> TIME_SERIES_TYPE:
    """
    Selects and returns the first of the values that tick (are modified) in the list provided.
    If more than one input is modified in the engine-cycle, it will return the first one that ticked in order of the
    list.
    """
    return next(tsl.modified_values()).delta_value


@graph(overloads=len_)
def len_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE], _sz: type[SIZE] = AUTO_RESOLVE) -> TS[int]:
    return const(_sz.SIZE)


@compute_node
def tsl_to_tsd(tsl: TSL[REF[TIME_SERIES_TYPE], SIZE], keys: tuple[str, ...]) -> TSD[str, REF[TIME_SERIES_TYPE]]:
    """
    Converts a time series into a time series dictionary with the keys provided.
    """
    return {k: ts.value for k, ts in zip(keys, tsl) if ts.modified}


@compute_node(overloads=getitem_, requires=lambda m, s: 0 <= s['index'] < m[SIZE])
def tsl_get_item_default(ts: REF[TSL[TIME_SERIES_TYPE, SIZE]], key: int) -> REF[TIME_SERIES_TYPE]:
    """
    Return a reference to an item in the TSL referenced
    """
    if ts.value.valid:
        if ts.value.has_peer:
            return PythonTimeSeriesReference(ts.value.output[key])
        else:
            return ts.value.items[key]
    else:
        return PythonTimeSeriesReference()


@compute_node(overloads=getitem_)
def tsl_get_item_ts(ts: REF[TSL[TIME_SERIES_TYPE, SIZE]], key: TS[int], _sz: Type[SIZE] = AUTO_RESOLVE) -> REF[TIME_SERIES_TYPE]:
    """
    Return a reference to an item in the TSL referenced
    """
    if key.value < 0 or key.value >= _sz.SIZE:
        return PythonTimeSeriesReference()

    if ts.value.valid:
        if ts.value.has_peer:
            return PythonTimeSeriesReference(ts.value.output[key.value])
        else:
            return ts.value.items[key.value]
    else:
        return PythonTimeSeriesReference()


@compute_node
def index_of(tsl: TSL[TIME_SERIES_TYPE, SIZE], ts: TIME_SERIES_TYPE) -> TS[int]:
    """
    Return the index of the leftmost time-series with the equal value to ts in the TSL
    """
    return next((i for i, t in enumerate(tsl) if t.valid and t.value == ts.value), -1)


@graph(overloads=sum_)
def sum_tsl_unary(tsl: TSL[TS[NUMBER], SIZE], tp: Type[TS[NUMBER]] = AUTO_RESOLVE) -> TS[NUMBER]:
    if len(tsl) == 1:
        return tsl[0]
    else:
        return _sum_tsl_unary(tsl, zero(tp, sum_))


@compute_node(overloads=sum_)
def _sum_tsl_unary(tsl: TSL[TS[NUMBER], SIZE], zero_ts: TS[NUMBER]) -> TS[NUMBER]:
    return sum((t.value for t in tsl.valid_values()), start=zero_ts.value)


@graph(overloads=sum_)
def sum_tsl_multi(*tsl: TSL[TSL[TIME_SERIES_TYPE, SIZE], SIZE_1]) -> TSL[TIME_SERIES_TYPE, SIZE_1]:
    """
    Item-wise sum() of the TSL elements. Missing elements on either side will cause a gap in the output
    """
    if len(tsl) == 1:
        return tsl[0]
    else:
        return TSL.from_ts(*(sum_(*tsls) for tsls in tsl))


@graph(overloads=add_)
def add_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise addition of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a + b for a, b in zip(lhs, rhs)))


@graph(overloads=sub_)
def sub_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise subtraction of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a - b for a, b in zip(lhs, rhs)))


@graph(overloads=mul_)
def mul_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise product of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a * b for a, b in zip(lhs, rhs)))


@graph(overloads=div_)
def div_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise division of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a / b for a, b in zip(lhs, rhs)))


@graph(overloads=floordiv_)
def floordiv_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise floor division of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a // b for a, b in zip(lhs, rhs)))


@graph(overloads=mod_)
def mod_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise lhs % rhs of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a % b for a, b in zip(lhs, rhs)))


@graph(overloads=pow_)
def pow_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise lhs ** rhs of the elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a ** b for a, b in zip(lhs, rhs)))


@graph(overloads=lshift_)
def lshift_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise leftshift of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a << b for a, b in zip(lhs, rhs)))


@graph(overloads=rshift_)
def rshift_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise leftshift of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a >> b for a, b in zip(lhs, rhs)))


@graph(overloads=bit_and)
def bit_and_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise bitwise AND of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a & b for a, b in zip(lhs, rhs)))


@graph(overloads=bit_or)
def bit_or_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise bitwise OR of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a | b for a, b in zip(lhs, rhs)))


@graph(overloads=bit_xor)
def bit_xor_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise bitwise XOR of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a ^ b for a, b in zip(lhs, rhs)))


@graph(overloads=eq_)
def eq_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TS[bool]:
    """
    Equality operator for two TSLs.  Asymmetrical missing values cause a False return
    """
    return len(lhs) == len(rhs) and all_(*(eq_(a, b) for a, b in zip(lhs, rhs)))


@graph(overloads=ne_)
def ne_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TS[bool]:
    """
    Not Equality operator for two TSLs.  Asymmetrical missing values cause a True return.
    """
    return not_(eq_(lhs, rhs))


@graph(overloads=neg_)
def neg_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise unary negative of the TSL elements
    """
    return TSL.from_ts(*(neg_(a) for a in ts))


@graph(overloads=pos_)
def pos_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise unary positive of the TSL elements
    """
    return TSL.from_ts(*(pos_(a) for a in ts))


@graph(overloads=invert_)
def invert_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise unary invert of the TSL elements
    """
    return TSL.from_ts(*(invert_(a) for a in ts))


@graph(overloads=abs_)
def abs_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise abs() of the TSL elements
    """
    return TSL.from_ts(*(abs_(a) for a in ts))


@graph(overloads=min_)
def min_tsl_unary(tsl: TSL[TIME_SERIES_TYPE, SIZE], tp: Type[TIME_SERIES_TYPE] = AUTO_RESOLVE) -> TIME_SERIES_TYPE:
    """
    Minimum value in the TSL
    """
    if len(tsl) == 1:
        return tsl[0]
    elif len(tsl) == 2:
        return min_(tsl[0], tsl[1])
    else:
        return reduce(lambda a, b: min_(a,  b), tsl, zero(tp, min_))


@graph(overloads=min_)
def min_tsl_multi(*tsl: TSL[TSL[TIME_SERIES_TYPE, SIZE], SIZE_1]) -> TSL[TIME_SERIES_TYPE, SIZE_1]:
    """
    Item-wise min() of the TSL elements. Missing elements on either side will cause a gap in the output
    """
    return TSL.from_ts(*(min_(*tsls) for tsls in tsl))


@graph(overloads=max_)
def max_tsl_unary(tsl: TSL[TIME_SERIES_TYPE, SIZE], tp: Type[TIME_SERIES_TYPE] = AUTO_RESOLVE) -> TIME_SERIES_TYPE:
    """
    Maximum value in the TSL
    """
    if len(tsl) == 1:
        return tsl[0]
    elif len(tsl) == 2:
        return max_(tsl[0], tsl[1])
    else:
        return reduce(lambda a, b: max_(a,  b), tsl, zero(tp, max_))


@graph(overloads=max_)
def max_tsl_multi(*tsl: TSL[TSL[TIME_SERIES_TYPE, SIZE], SIZE_1]) -> TSL[TIME_SERIES_TYPE, SIZE_1]:
    """
    Item-wise max() of the TSL elements. Missing elements on either side will cause a gap in the output
    """
    return TSL.from_ts(*(max_(*tsls) for tsls in tsl))


@compute_node(overloads=str_)
def str_(ts: TSL[TIME_SERIES_TYPE, SIZE], tp: Type[TIME_SERIES_TYPE] = AUTO_RESOLVE) -> TS[str]:
    return str(ts.value)


@compute_node(valid=tuple(), overloads=union)
def union_tsl_tss(*tsl: TSL[TSS[KEYABLE_SCALAR], SIZE], _output: TSS_OUT[KEYABLE_SCALAR] = None) -> TSS[KEYABLE_SCALAR]:
    tss: TSS[KEYABLE_SCALAR, SIZE]
    to_add: set[KEYABLE_SCALAR] = set()
    to_remove: set[KEYABLE_SCALAR] = set()
    for tss in tsl.modified_values():
        to_add |= tss.added()
        to_remove |= tss.removed()
    if (disputed := to_add.intersection(to_remove)):
        # These items are marked for addition and removal, so at least some set is hoping to add these items.
        # Thus, overall these are an add, unless they are already added.
        new_items = disputed.intersection(_output.value)
        to_remove -= new_items
    to_remove &= _output.value  # Only remove items that are already in the output.
    if to_remove:
        # Now we need to make sure there are no items that may be duplicated in other inputs.
        for tss in tsl.valid_values():
            to_remove -= to_remove.intersection(tss.value)  # Remove items that exist in an input
            if not to_remove:
                break
    return PythonSetDelta(to_add, to_remove)

==> hgraph-0.2.43-py3-none-any.whl: hgraph/nodes/_tsl_operators.py <==
from statistics import stdev, variance
from typing import Type

from hgraph import (
    compute_node,
    TSL,
    TIME_SERIES_TYPE,
    SIZE,
    SCALAR,
    TS,
    graph,
    AUTO_RESOLVE,
    NUMBER,
    REF,
    TSD,
    union,
    TSS,
    KEYABLE_SCALAR,
    TSS_OUT,
    PythonSetDelta,
    add_,
    sub_,
    mul_,
    div_,
    floordiv_,
    mod_,
    pow_,
    lshift_,
    rshift_,
    bit_and,
    bit_or,
    bit_xor,
    eq_,
    ne_,
    not_,
    neg_,
    pos_,
    invert_,
    abs_,
    min_,
    max_,
    reduce,
    zero,
    str_,
    PythonTimeSeriesReference,
    len_,
    sum_,
    getitem_,
    all_,
    clone_typevar,
    mean,
    std,
    var,
)

__all__ = ("flatten_tsl_values", "tsl_to_tsd", "index_of")


@compute_node
def flatten_tsl_values(tsl: TSL[TIME_SERIES_TYPE, SIZE], all_valid: bool = False) -> TS[tuple[SCALAR, ...]]:
    """
    This will convert the TSL into a time-series of tuples. The value will be the value type of the time-series
    provided to the TSL. If the value has not ticked yet, the value will be None.
    The output type must be defined by the user.

    Usage:
    ```Python
        tsl: TSL[TS[float], Size[3]] = ...
        out = flatten_tsl[SCALAR: float](tsl)
    ```
    """
    return tsl.value if not all_valid or tsl.all_valid else None


@graph(overloads=len_)
def len_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE], _sz: type[SIZE] = AUTO_RESOLVE) -> TS[int]:
    from hgraph import const

    return const(_sz.SIZE)


@compute_node(deprecated="Use combine(keys, tsl)")
def tsl_to_tsd(tsl: TSL[REF[TIME_SERIES_TYPE], SIZE], keys: tuple[str, ...]) -> TSD[str, REF[TIME_SERIES_TYPE]]:
    """
    Converts a time series into a time series dictionary with the keys provided.
    """
    return {k: ts.value for k, ts in zip(keys, tsl) if ts.modified}


@compute_node(overloads=getitem_, requires=lambda m, s: 0 <= s["index"] < m[SIZE])
def getitem_tsl_scalar(ts: REF[TSL[TIME_SERIES_TYPE, SIZE]], key: int) -> REF[TIME_SERIES_TYPE]:
    """
    Return a reference to an item in the TSL referenced
    """
    if ts.value.valid:
        if ts.value.has_peer:
            return PythonTimeSeriesReference(ts.value.output[key])
        else:
            return ts.value.items[key]
    else:
        return PythonTimeSeriesReference()


@compute_node(overloads=getitem_)
def getitem_tsl_ts(
    ts: REF[TSL[TIME_SERIES_TYPE, SIZE]], key: TS[int], _sz: Type[SIZE] = AUTO_RESOLVE
) -> REF[TIME_SERIES_TYPE]:
    """
    Return a reference to an item in the TSL referenced
    """
    if key.value < 0 or key.value >= _sz.SIZE:
        return PythonTimeSeriesReference()

    if ts.value.valid:
        if ts.value.has_peer:
            return PythonTimeSeriesReference(ts.value.output[key.value])
        else:
            return ts.value.items[key.value]
    else:
        return PythonTimeSeriesReference()


@compute_node
def index_of(tsl: TSL[TIME_SERIES_TYPE, SIZE], ts: TIME_SERIES_TYPE) -> TS[int]:
    """
    Return the index of the leftmost time-series in the TSL with value equal to ts
    """
    return next((i for i, t in enumerate(tsl) if t.valid and t.value == ts.value), -1)


@graph(overloads=sum_)
def sum_tsl_unary(tsl: TSL[TS[NUMBER], SIZE], tp: Type[TS[NUMBER]] = AUTO_RESOLVE) -> TS[NUMBER]:
    if len(tsl) == 1:
        return tsl[0]
    else:
        return _sum_tsl_unary(tsl, zero(tp, sum_))


@compute_node(overloads=sum_)
def _sum_tsl_unary(tsl: TSL[TS[NUMBER], SIZE], zero_ts: TS[NUMBER]) -> TS[NUMBER]:
    return sum((t.value for t in tsl.valid_values()), start=zero_ts.value)


SIZE_1 = clone_typevar(SIZE, "SIZE_1")


@graph(overloads=sum_)
def sum_tsl_multi(*tsl: TSL[TSL[TIME_SERIES_TYPE, SIZE], SIZE_1]) -> TSL[TIME_SERIES_TYPE, SIZE_1]:
    """
    Item-wise sum() of the TSL elements. Missing elements on either side will cause a gap in the output
    """
    if len(tsl) == 1:
        return tsl[0]
    else:
        return TSL.from_ts(*(sum_(*tsls) for tsls in tsl))


@graph(overloads=add_)
def add_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise addition of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a + b for a, b in zip(lhs, rhs)))


@graph(overloads=sub_)
def sub_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise subtraction of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a - b for a, b in zip(lhs, rhs)))


@graph(overloads=mul_)
def mul_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise product of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a * b for a, b in zip(lhs, rhs)))


@graph(overloads=div_)
def div_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise division of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a / b for a, b in zip(lhs, rhs)))


@graph(overloads=floordiv_)
def floordiv_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise floor division of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a // b for a, b in zip(lhs, rhs)))


@graph(overloads=mod_)
def mod_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise lhs % rhs of TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a % b for a, b in zip(lhs, rhs)))


@graph(overloads=pow_)
def pow_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise lhs ** rhs of the elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a**b for a, b in zip(lhs, rhs)))


@graph(overloads=lshift_)
def lshift_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise leftshift of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a << b for a, b in zip(lhs, rhs)))


@graph(overloads=rshift_)
def rshift_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise leftshift of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a >> b for a, b in zip(lhs, rhs)))


@graph(overloads=bit_and)
def bit_and_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise bitwise AND of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a & b for a, b in zip(lhs, rhs)))


@graph(overloads=bit_or)
def bit_or_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise bitwise OR of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a | b for a, b in zip(lhs, rhs)))


@graph(overloads=bit_xor)
def bit_xor_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise bitwise XOR of the TSL elements.  A missing value on either lhs or rhs causes a gap on the output
    """
    return TSL.from_ts(*(a ^ b for a, b in zip(lhs, rhs)))


@graph(overloads=eq_)
def eq_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TS[bool]:
    """
    Equality operator for two TSLs.  Asymmetrical missing values cause a False return
    """
    return len(lhs) == len(rhs) and all_(*(eq_(a, b) for a, b in zip(lhs, rhs)))


@graph(overloads=ne_)
def ne_tsls(lhs: TSL[TIME_SERIES_TYPE, SIZE], rhs: TSL[TIME_SERIES_TYPE, SIZE]) -> TS[bool]:
    """
    Not Equality operator for two TSLs.  Asymmetrical missing values cause a True return.
    """
    return not_(eq_(lhs, rhs))


@graph(overloads=neg_)
def neg_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise unary negative of the TSL elements
    """
    return TSL.from_ts(*(neg_(a) for a in ts))


@graph(overloads=pos_)
def pos_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise unary positive of the TSL elements
    """
    return TSL.from_ts(*(pos_(a) for a in ts))


@graph(overloads=invert_)
def invert_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise unary invert of the TSL elements
    """
    return TSL.from_ts(*(invert_(a) for a in ts))


@graph(overloads=abs_)
def abs_tsl(ts: TSL[TIME_SERIES_TYPE, SIZE]) -> TSL[TIME_SERIES_TYPE, SIZE]:
    """
    Item-wise abs() of the TSL elements
    """
    return TSL.from_ts(*(abs_(a) for a in ts))


@graph(overloads=min_)
def min_tsl_unary(tsl: TSL[TIME_SERIES_TYPE, SIZE], tp: Type[TIME_SERIES_TYPE] = AUTO_RESOLVE) -> TIME_SERIES_TYPE:
    """
    Minimum value in the TSL
    """
    if len(tsl) == 1:
        return tsl[0]
    elif len(tsl) == 2:
        return min_(tsl[0], tsl[1])
    else:
        return reduce(lambda a, b: min_(a, b), tsl, zero(tp, min_))


@graph(overloads=min_)
def min_tsl_multi(*tsl: TSL[TSL[TIME_SERIES_TYPE, SIZE], SIZE_1]) -> TSL[TIME_SERIES_TYPE, SIZE_1]:
    """
    Item-wise min() of the TSL elements. Missing elements on either side will cause a gap in the output
    """
    return TSL.from_ts(*(min_(*tsls) for tsls in tsl))


@graph(overloads=max_)
def max_tsl_unary(tsl: TSL[TIME_SERIES_TYPE, SIZE], tp: Type[TIME_SERIES_TYPE] = AUTO_RESOLVE) -> TIME_SERIES_TYPE:
    """
    Maximum value in the TSL
    """
    if len(tsl) == 1:
        return tsl[0]
    elif len(tsl) == 2:
        return max_(tsl[0], tsl[1])
    else:
        return reduce(lambda a, b: max_(a, b), tsl, zero(tp, max_))


@graph(overloads=max_)
def max_tsl_multi(*tsl: TSL[TSL[TIME_SERIES_TYPE, SIZE], SIZE_1]) -> TSL[TIME_SERIES_TYPE, SIZE_1]:
    """
    Item-wise max() of the TSL elements. Missing elements on either side will cause a gap in the output
    """
    return TSL.from_ts(*(max_(*tsls) for tsls in tsl))


@compute_node(overloads=str_)
def str_(ts: TSL[TIME_SERIES_TYPE, SIZE], tp: Type[TIME_SERIES_TYPE] = AUTO_RESOLVE) -> TS[str]:
    return str(ts.value)


@compute_node(valid=tuple(), overloads=union)
def union_tsl_tss(*tsl: TSL[TSS[KEYABLE_SCALAR], SIZE], _output: TSS_OUT[KEYABLE_SCALAR] = None) -> TSS[KEYABLE_SCALAR]:
    tss: TSS[KEYABLE_SCALAR, SIZE]
    to_add: set[KEYABLE_SCALAR] = set()
    to_remove: set[KEYABLE_SCALAR] = set()
    for tss in tsl.modified_values():
        to_add |= tss.added()
        to_remove |= tss.removed()
    if disputed := to_add.intersection(to_remove):
        # These items are marked for addition and removal, so at least some set is hoping to add these items.
        # Thus, overall these are an add, unless they are already added.
        new_items = disputed.intersection(_output.value)
        to_remove -= new_items
    to_remove &= _output.value  # Only remove items that are already in the output.
    if to_remove:
        # Now we need to make sure there are no items that may be duplicated in other inputs.
        for tss in tsl.valid_values():
            to_remove -= to_remove.intersection(tss.value)  # Remove items that exist in an input
            if not to_remove:
                break
    return PythonSetDelta(to_add, to_remove)


@graph(overloads=mean)
def mean_tsl_multi(*tsl: TSL[TSL[TIME_SERIES_TYPE, SIZE], SIZE_1]) -> TSL[TS[float], SIZE_1]:
    """
    Item-wise mean() of the TSL elements. Missing elements on either side will cause a gap in the output
    """
    if len(tsl) == 1:
        return tsl[0]
    else:
        return TSL.from_ts(*(mean(*tsls) for tsls in tsl))


@graph(overloads=mean)
def mean_tsl_unary_number(ts: TSL[TS[NUMBER], SIZE], _sz: type[SIZE] = AUTO_RESOLVE) -> TS[float]:
    from hgraph import DivideByZero

    return div_(sum_(ts), _sz.SIZE, divide_by_zero=DivideByZero.NAN)


@graph(overloads=std)
def std_tsl_multi(*tsl: TSL[TSL[TIME_SERIES_TYPE, SIZE], SIZE_1]) -> TSL[TS[float], SIZE_1]:
    """
    Item-wise std() of the TSL elements. Missing elements on either side will cause a gap in the output
    """
    if len(tsl) == 1:
        return std(tsl[0])
    else:
        return TSL.from_ts(*(std(*tsls) for tsls in tsl))


@compute_node(overloads=std)
def std_tsl_unary_number(ts: TSL[TS[NUMBER], SIZE], _sz: type[SIZE] = AUTO_RESOLVE) -> TS[float]:
    valid_elements = tuple(t.value for t in ts if t.valid)
    n_valid = len(valid_elements)
    if n_valid <= 1:
        return 0.0
    else:
        return stdev(valid_elements)


@graph(overloads=var)
def var_tsl_multi(*tsl: TSL[TSL[TIME_SERIES_TYPE, SIZE], SIZE_1]) -> TSL[TS[float], SIZE_1]:
    """
    Item-wise std() of the TSL elements. Missing elements on either side will cause a gap in the output
    """
    if len(tsl) == 1:
        return var(tsl[0])
    else:
        return TSL.from_ts(*(var(*tsls) for tsls in tsl))


@compute_node(overloads=var)
def var_tsl_unary_number(ts: TSL[TS[NUMBER], SIZE], _sz: type[SIZE] = AUTO_RESOLVE) -> TS[float]:
    valid_elements = tuple(t.value for t in ts if t.valid)
    n_valid = len(valid_elements)
    if n_valid <= 1:
        return 0.0
    else:
        return float(variance(valid_elements))
